#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from .tool.communication.lazycache import Cache
from .tool.communication.report import Report
from .tool.dataflow.autoins import AutoIns
from .tool.enhancement.attribute.binarize import Binarize
from .tool.enhancement.attribute.pca import pca
from .tool.evaluation.partition import Partition
from .tool.evaluation.split import split
from .tool.evaluation.summ import Summ2
from .tool.learning.supervised.classification.svm import svm
from .tool.stream.map import Map
from .tool.stream.reduce import Reduce

from .tool.evaluation.metric import metric
from .tool.enhancement.attribute.binarize import binarize



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from pprint import pprint

from akangatu.container import Container1, ContainerN


def set_leaf(tree, branches, leaf):
    """ Set a terminal element to *leaf* within nested dictionaries.
    *branches* defines the path through dictionnaries.

    Example:
    >>> t = {}
    >>> set_leaf(t, ['b1','b2','b3'], 'new_leaf')
    >>> print(t)
    {'b1': {'b2': {'b3': 'new_leaf'}}}

    https://stackoverflow.com/a/12389499/9681577
    """
    if len(branches) == 1:
        tree[branches[0]] = leaf
        return
    if branches[0] not in tree:
        tree[branches[0]] = {}
    set_leaf(tree[branches[0]], branches[1:], leaf)


def tools_asdict():
    import importlib
    from kururu import tool
    import sys
    import os
    from pathlib import Path
    m = {}
    path = os.path.abspath(tool.__file__)
    startpath = "/".join(path.split("/")[:-1])
    paths = list(Path(startpath).rglob("*.py"))
    for f in paths:
        if not (f.name.startswith("_") or "/stream/" in str(f) or "/abs/" in str(f) or "/mixin/" in str(f)):
            sys.path.append("/".join(str(f).split("/")[:-1]))
            arq = f.name[:-3]
            a = importlib.import_module(arq)
            for k, v in a.__dict__.items():
                if arq == k.lower() and not k.islower():
                    if not issubclass(v, (Container1, ContainerN)):
                        step = v()
                        m[f.name] = {"step": step.desc, "params": step.parameters}

    tree = {}
    for root, dirs, files in os.walk(startpath):
        branches = [startpath]
        if root != startpath:
            relpath = os.path.relpath(root, startpath)
            branches.extend(relpath.split('/'))

        set_leaf(tree, branches, dict([(d, {}) for d in dirs] + [(f, None) for f in files]))

    def clean(tr):
        dic = {}
        if isinstance(tr, dict):
            dic = {}
            for k, v in tr.items():
                if not (k.startswith("_") and k not in ["stream", "abs", "mixin"]):
                    dic[k] = m[k] if k in m else clean(v)
        return dic

    return clean(list(tree.values())[0])


pprint(tools_asdict(), depth=5)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

# from matplotlib import pyplot as plt
#
# from akangatu.ddstep import DDStep
# from melime.explainers.explainer import Explainer
# from melime.explainers.visualizations.plot_importance import ExplainGraph
# from melime.generators.kde_gen import KDEGen
# # from akangatu.transf.mixin.noop import asNoOp
#
#
# # TODO: simplify this class to be independent of the prediction of syntetic examples.
# class Explain(asNoOp, DDStep):
#     def __init__(self, inner=None, class_index=0):
#     raise NotImplemented
#     super().__init__(inner, {"class_index": class_index})
#     # TODO: dar opção de selecionar a classe mais forte, mas depende de passar o modelo (issue no melime?)
#     self.class_index = class_index
#
#     def _process_(self, data):
#     inner = data.inner
#     generator = KDEGen(verbose=True).fit(inner.X)
#     # TODO: model_predict depende do modelo pra predizer em exemplos artificiais
#     explainer = Explainer(
#         model_predict=??????????????????????????????,
#         generator=generator,
#         local_model='Ridge',
#         feature_names=data.Xd,
#         target_names=data.Yd
#     )
#     explanation, counterfactual_examples = explainer.explain_instance(
#         x_explain=data.X[0:1],
#         class_index=self.class_index,
#         r=1.0,
#         n_samples=500,
#         tol_importance=0.1,
#         tol_error=0.1,
#         local_mini_batch_max=40,
#         scale_data=False,
#         weight_kernel='gaussian'
#     )
#     fig, axs = ExplainGraph.plot(explanation.explain())
#     # plt.savefig(f'iris_x_{",".join(data.Xd)}_m-lime_kde.svg')
#     plt.show()  # pip install PyQt5==5.12.2  # plotlib is currently using agg, which is a non-GUI backend, so canno
#     fig, axs = ExplainGraph.plot_errors(explanation)
#
#     return data



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from aiuna.content.data import Data
from akangatu.distep import DIStep
from akangatu.transf.mixin.noop import asNoOp
from tatu import Tatu
from tatu.abs.storage import Storage
from tatu.amnesia import Amnesia
from tatu.pickle_ import Pickle
from tatu.sql.sqlite import SQLite


def setcache(storage, eager_store=True):
    from akangatu.transf.config import CACHE
    CACHE["cache"] = Cache(storage, eager_store=eager_store)


def closecache():
    from akangatu.transf.config import CACHE
    CACHE["cache"].storage.close()


class Cache(asNoOp, DIStep):
    storages = {}

    #     ???? ver se vai haver cache-container.
    #     o único motivo seria caso a montagem dos campos lazy tenha muito overhead
    def __init__(self, storage="sqlite", eager_store=True, ignorelock=True, seed=0):  # TODO: what todo with seed here?
        DIStep.__init__(self, storage=storage, eager_store=eager_store, ignorelock=ignorelock, seed=seed)
        if storage == "pickle":
            storage = Pickle(close_when_idle=True)
        elif storage == "sqlite":
            storage = SQLite(close_when_idle=True)
        elif storage == "amnesia":
            storage = Amnesia(close_when_idle=True)
        elif isinstance(storage, Storage):
            pass
        elif "://" in storage:
            storage = Tatu(url=storage, close_when_idle=True)
        else:
            print("Unknown storage:", storage)
            exit()
        if storage.id not in self.storages:
            self.storages[storage.id] = storage
        self.storage: Storage = self.storages[storage.id]
        self.eager_store = eager_store
        self.ignorelock = ignorelock

    def _process_(self, data: Data):
        fetched = self.storage.fetch(data, lock=True, ignorelock=self.ignorelock)
        if fetched:
            return fetched
        self.storage.store(data, unlock=True, lazy=not self.eager_store)
        return data

    # TODO dar tatu.unlock() dentro do data.getitem se ocorrer alguma exception que interrompa o cacheamento

# c = 0



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from akangatu.distep import DIStep
from akangatu.transf.mixin.config import asConfigLess


class FailureAtStream(Exception):
    """At least one of the streamed Data objects failed during build."""


class Reduce(asConfigLess, DIStep):
    def _process_(self, data):
        # TODO: centralizar logs usando essas tecnologias :
        def consume():
            if data.stream is None:
                def debug_print(*arg):
                    from inspect import currentframe
                    frameinfo = currentframe()
                    print(frameinfo.f_back.f_code.co_filename, " Line:", frameinfo.f_back.f_lineno)
                    print(*arg)

                debug_print(
                    "\nHistory: " + str(data.history) + f"\n{self.name} needs a Data object containing a stream.")
                print("Missing stream inside", data.id)
                exit()

            failures = []  # TODO:checar se falhados estão sendo tratados/interrompidos no Step (caso geral)
            for d in data.stream:
                # print("consome", d.id)
                if d.failure:
                    print("Failure:", d.failure)
                    failures.append(d.failure)
            if failures:
                raise FailureAtStream(data.failed([], "; ".join(failures)))
            return

        def f(lazyval):
            def func():
                consume()
                return lazyval()

            return func

        dic = {}
        for k, v in data.field_funcs_m.items():
            if k.startswith("_"):
                dic[k[1:]] = f(v)
                dic[k] = None
            elif k == "stream":
                dic[k] = None

        return data.update(self, **dic)

    def translate(self, exception, data):
        if isinstance(exception, FailureAtStream):
            return f"{data.id} stream failures: {str(exception)}"

# class Reduce2(asMacro, Reduce):
#     def _step_(self):
#     return Reduce * In(Reduce)  # REMINDER: talvez um dia faça sentido



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#

from imblearn.over_sampling import RandomOverSampler as imbROS

from akangatu.abs.mixin.macro import asMacro
from akangatu.ddstep import DDStep
from akangatu.operator.unary.inop import In
from kururu.tool.enhancement.instance.sampling.abs.sampler import Sampler


class ROS_(Sampler):
    """Increase number of instances of all classes so as to be equal to the amount of instances in the majority class
    (default behavior)."""

    def __init__(self, strategy="not majority", seed=0):
        super().__init__(strategy=strategy, seed=seed)
        self.strategy = strategy
        self.seed = seed

    def _algorithm_func(self):
        return lambda: imbROS(sampling_strategy=self.strategy, random_state=self.seed)


# class ROS(DDStep):
#     """Apply random over sampling to process inner data instead of using the outer data."""
#
#     def __init__(self, inner=None, strategy="not majority", seed=0):
#         DDStep.__init__(self, inner, strategy=strategy, seed=seed)
#         self.strategy = strategy
#         self.seed = seed
#
#     def _process_(self, data):
#         d = data >> In(ROS_(**self.held))
#         return data.update(self, **{k: d.field_funcs_m[k] for k in d.changed})
#
class ROS(asMacro, DDStep):
    """Apply random over sampling to process inner data instead of using the outer data."""
    def __init__(self, inner=None, strategy="not majority", seed=0):
        DDStep.__init__(self, inner, strategy=strategy, seed=seed)
        self.strategy = strategy
        self.seed = seed

    def _step_(self):
        return In(ROS_(**self.held))
